/*
******************************************************************************
* File Name    : printfUart1_DMA_Driver.c
* Created On   : 4 Mar. 2022
* Description  :
* Author	     : R J Garnett - Admin
******************************************************************************
*/

#include "stm32h7xx_hal.h"
#include "circular_buffer_char.h"
#include "stdarg.h"
#include "stdint.h"
#include "printf-stdarg.h"
#include "stdbool.h"
#include "cmsis_os.h"
#include "FreeRTOS.h"

UART_HandleTypeDef huart1;
DMA_HandleTypeDef hdma_usart1_tx;
DMA_HandleTypeDef hdma_usart1_rx;

typedef enum
{
	PRINTF_OK,
	PRINTF_BUFFER_FULL,
	PRINTF_HAL_DRIVER_ERROR,
	PRINTF_INIT_FAILED,

}status_t_enum ;

#define VCP_RX_Pin GPIO_PIN_10
#define VCP_RX_GPIO_Port GPIOA
#define VCP_TX_Pin GPIO_PIN_9
#define VCP_TX_GPIO_Port GPIOA

#define DMA_BUFF_TX_SIZE 512
#define DMA_BUFF_RX_SIZE 512

#define CIRC_BUFF_TX_SIZE (DMA_BUFF_TX_SIZE * 5)
#define CIRC_BUFF_RX_SIZE (DMA_BUFF_TX_SIZE * 2)

#define TEMP_BUFF_SIZE 133

char uart1DMABuffTx[DMA_BUFF_TX_SIZE] __attribute__((section(".UART1_TX_DMA_RAM")));
char uart1DMABuffRx[DMA_BUFF_RX_SIZE] __attribute__((section(".UART1_RX_DMA_RAM")));


/************** CIRCULAR BUFFER ***************/

cbuf_handle_t  circBuffTxHandle;
bool circBufferOverunFlag = false;

char __attribute__((section (".EXT_RAM_STATS_OUT"))) txCircBuff[CIRC_BUFF_TX_SIZE];
char tempBuff[TEMP_BUFF_SIZE];


static void MX_USART1_UART_Init(void);
HAL_StatusTypeDef HAL_UART_1_Init(UART_HandleTypeDef *huart);
HAL_StatusTypeDef HAL_UART_Transmit_DMA_rjg(UART_HandleTypeDef *huart, uint8_t *pData, uint16_t Size);
static void Error_Handler(void);
static int stringLength_n(const char* buffer, int sizeBuff);
static bool startPrintTask(void);
static void printTask(void * argument);

/* Static FreeRTOS Variables */
#define STACK_SIZE_START_WIFI_TASK 1024

TaskHandle_t 	printTaskHandle;
StackType_t 	printTaskStack[STACK_SIZE_START_WIFI_TASK];
StaticTask_t 	printTaskBuffer;



/*************************************************************************************/
void initPrintfUart1(void)
{
	/* Initialise the circular buffer handle. */
  circular_buf_char_init(&circBuffTxHandle, (char*)txCircBuff, CIRC_BUFF_TX_SIZE);

  /* Initialize uart1 which is connected to the ST Link on board debugger */
  MX_USART1_UART_Init();

  /* Start the printing thread */
  startPrintTask();
}



/*************************************************************************************/
int printfdma(const char *format, ...)
{
	int nResult = 0;
	int k;
  va_list args;

  if(!circular_buf_char_full(&circBuffTxHandle))
  {
		va_start( args, format);

		nResult = sprintf(&tempBuff[0], format, args);

		if (stringLength_n(tempBuff, TEMP_BUFF_SIZE) > TEMP_BUFF_SIZE - 1)
		{
			tempBuff[TEMP_BUFF_SIZE - 1] = 0x0;
			nResult = 132;
		}

		/* put data in circular buffer. */
		for(k = 0; k < nResult; k++)
		{
			circBufferOverunFlag = !circular_buf_char_try_put(&circBuffTxHandle, tempBuff[k]);
		}

		/* Kick off the uart */
		xTaskNotifyGive(printTaskHandle);
  }
  else
  {
  	/* Circular buffer full - post message TODO */
  }
  return nResult;
}


/*
 ***************************************************************************
 * HELPER FUNCTIONS
 *
 ***************************************************************************
 */
static int stringLength_n(const char* buffer, int sizeBuff)
{
	for (int i=0; sizeBuff; i++)
	{
		if(buffer[i] == 0x0)
			{
				return i - 1;
			}
	}
	return -1;
}

/*
 ***************************************************************************
 * FREE RTOS TASKS AND SEMAPHORE
 *
 ***************************************************************************
 */


static bool startPrintTask(void)
{
	printTaskHandle 	 = xTaskCreateStatic(printTask,
																				 "printTask",
																				 STACK_SIZE_START_WIFI_TASK,
																				 NULL,
																				 osPriorityBelowNormal3,
																				 printTaskStack,
																				 &printTaskBuffer);

	if(printTaskHandle == NULL)
	{
		return PRINTF_INIT_FAILED;
	}
	return PRINTF_OK;
}


/*************************************************************************************/
static void printTask(void * argument)
{
	static uint32_t thread_notification;
	static int k = 0;
	char chr[1];

	/* call the thread notification to set the counter to zero so it blocks */
	thread_notification = ulTaskNotifyTake(pdTRUE, 0);

	while(1)
	{
    thread_notification = ulTaskNotifyTake(pdTRUE, portMAX_DELAY);
    if(thread_notification)
    {
			while(!circular_buf_char_empty(&circBuffTxHandle) && k < DMA_BUFF_TX_SIZE)
			{
				circular_buf_char_pop(&circBuffTxHandle, (char*)&chr);
				uart1DMABuffTx[k] = *chr;
				k++;
			}

			k--;
			if( HAL_UART_Transmit_DMA_rjg(&huart1,  (unsigned char*)&uart1DMABuffTx, k) != HAL_OK)
			{
				/* TODO Hal error handling */
			}
		} /* end of thread notification true */
	}
	vTaskDelay(5);
}



/*
 ***************************************************************************
 * UART IO AND INTERRUPT FUNCTIONS
 *
 ***************************************************************************
 */

/**
  * @brief USART1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_USART1_UART_Init(void)
{
  huart1.Instance = USART1;
  huart1.Init.BaudRate = 115200;
  huart1.Init.WordLength = UART_WORDLENGTH_8B;
  huart1.Init.StopBits = UART_STOPBITS_1;
  huart1.Init.Parity = UART_PARITY_NONE;
  huart1.Init.Mode = UART_MODE_TX_RX;
  huart1.Init.HwFlowCtl = UART_HWCONTROL_NONE;
  huart1.Init.OverSampling = UART_OVERSAMPLING_16;
  huart1.Init.OneBitSampling = UART_ONE_BIT_SAMPLE_DISABLE;
  huart1.Init.ClockPrescaler = UART_PRESCALER_DIV1;
  huart1.AdvancedInit.AdvFeatureInit = UART_ADVFEATURE_NO_INIT;
  if (HAL_UART_Init(&huart1) != HAL_OK)
  {
    Error_Handler();
  }
  if (HAL_UARTEx_SetTxFifoThreshold(&huart1, UART_TXFIFO_THRESHOLD_1_8) != HAL_OK)
  {
    Error_Handler();
  }
  if (HAL_UARTEx_SetRxFifoThreshold(&huart1, UART_RXFIFO_THRESHOLD_1_8) != HAL_OK)
  {
    Error_Handler();
  }
  if (HAL_UARTEx_EnableFifoMode(&huart1) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
* @brief UART MSP Initialization
* This function configures the hardware resources used in this example
* @param huart: UART handle pointer
* @retval None
*/
static void HAL_UART_1_MspInit(UART_HandleTypeDef* huart)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};
  RCC_PeriphCLKInitTypeDef PeriphClkInitStruct = {0};

  /** Initializes the peripherals clock
  */
	PeriphClkInitStruct.PeriphClockSelection = RCC_PERIPHCLK_USART1;
	PeriphClkInitStruct.Usart16ClockSelection = RCC_USART16910CLKSOURCE_D2PCLK2;
	if (HAL_RCCEx_PeriphCLKConfig(&PeriphClkInitStruct) != HAL_OK)
	{
		Error_Handler();
	}

	/* Peripheral clock enable */
	__HAL_RCC_USART1_CLK_ENABLE();
	__HAL_RCC_GPIOA_CLK_ENABLE();
	/**USART1 GPIO Configuration
	PA10     ------> USART1_RX
	PA9     ------> USART1_TX
	*/
	GPIO_InitStruct.Pin = VCP_RX_Pin|VCP_TX_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_AF_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
	GPIO_InitStruct.Alternate = GPIO_AF7_USART1;
	HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

	/* USART1 DMA Init */
	/* USART1_TX Init */
	hdma_usart1_tx.Instance = DMA1_Stream7;
	hdma_usart1_tx.Init.Request = DMA_REQUEST_USART1_TX;
	hdma_usart1_tx.Init.Direction = DMA_MEMORY_TO_PERIPH;
	hdma_usart1_tx.Init.PeriphInc = DMA_PINC_DISABLE;
	hdma_usart1_tx.Init.MemInc = DMA_MINC_ENABLE;
	hdma_usart1_tx.Init.PeriphDataAlignment = DMA_PDATAALIGN_BYTE;
	hdma_usart1_tx.Init.MemDataAlignment = DMA_MDATAALIGN_BYTE;
	hdma_usart1_tx.Init.Mode = DMA_NORMAL;
	hdma_usart1_tx.Init.Priority = DMA_PRIORITY_LOW;
	hdma_usart1_tx.Init.FIFOMode = DMA_FIFOMODE_DISABLE;
	if (HAL_DMA_Init(&hdma_usart1_tx) != HAL_OK)
	{
		Error_Handler();
	}

	__HAL_LINKDMA(huart,hdmatx,hdma_usart1_tx);

	/* USART1_RX Init */
	hdma_usart1_rx.Instance = DMA1_Stream6;
	hdma_usart1_rx.Init.Request = DMA_REQUEST_USART1_RX;
	hdma_usart1_rx.Init.Direction = DMA_PERIPH_TO_MEMORY;
	hdma_usart1_rx.Init.PeriphInc = DMA_PINC_DISABLE;
	hdma_usart1_rx.Init.MemInc = DMA_MINC_ENABLE;
	hdma_usart1_rx.Init.PeriphDataAlignment = DMA_PDATAALIGN_BYTE;
	hdma_usart1_rx.Init.MemDataAlignment = DMA_MDATAALIGN_BYTE;
	hdma_usart1_rx.Init.Mode = DMA_NORMAL;
	hdma_usart1_rx.Init.Priority = DMA_PRIORITY_LOW;
	hdma_usart1_rx.Init.FIFOMode = DMA_FIFOMODE_DISABLE;
	if (HAL_DMA_Init(&hdma_usart1_rx) != HAL_OK)
	{
		Error_Handler();
	}

	__HAL_LINKDMA(huart,hdmarx,hdma_usart1_rx);

	/* USART1 interrupt Init */
	HAL_NVIC_SetPriority(USART1_IRQn, 10, 0);
	HAL_NVIC_EnableIRQ(USART1_IRQn);

  __HAL_RCC_DMA1_CLK_ENABLE();

  /* DMA1_Stream0_IRQn interrupt configuration */
  HAL_NVIC_SetPriority(DMA1_Stream7_IRQn, 8, 0);
  HAL_NVIC_EnableIRQ(DMA1_Stream7_IRQn);

  /* DMA1_Stream1_IRQn interrupt configuration */
  HAL_NVIC_SetPriority(DMA1_Stream6_IRQn, 8, 0);
  HAL_NVIC_EnableIRQ(DMA1_Stream6_IRQn);
}


/**
  * @brief Initialize the UART mode according to the specified
  *        parameters in the UART_InitTypeDef and initialize the associated handle.
  * @param huart UART handle.
  * @retval HAL status
  */
HAL_StatusTypeDef HAL_UART_1_Init(UART_HandleTypeDef *huart)
{
  /* Check the UART handle allocation */
  if (huart == NULL)
  {
    return HAL_ERROR;
  }

  if (huart->Init.HwFlowCtl != UART_HWCONTROL_NONE)
  {
    /* Check the parameters */
    assert_param(IS_UART_HWFLOW_INSTANCE(huart->Instance));
  }
  else
  {
    /* Check the parameters */
    assert_param((IS_UART_INSTANCE(huart->Instance)) || (IS_LPUART_INSTANCE(huart->Instance)));
  }

  if (huart->gState == HAL_UART_STATE_RESET)
  {
    /* Allocate lock resource and initialize it */
    huart->Lock = HAL_UNLOCKED;

#if (USE_HAL_UART_REGISTER_CALLBACKS == 1)
    UART_InitCallbacksToDefault(huart);

    if (huart->MspInitCallback == NULL)
    {
      huart->MspInitCallback = HAL_UART_MspInit;
    }

    /* Init the low level hardware */
    huart->MspInitCallback(huart);
#else
    /* Init the low level hardware : GPIO, CLOCK */
    HAL_UART_1_MspInit(huart);
#endif /* (USE_HAL_UART_REGISTER_CALLBACKS) */
  }

  huart->gState = HAL_UART_STATE_BUSY;

  __HAL_UART_DISABLE(huart);

  /* Set the UART Communication parameters */
  if (UART_SetConfig(huart) == HAL_ERROR)
  {
    return HAL_ERROR;
  }

  if (huart->AdvancedInit.AdvFeatureInit != UART_ADVFEATURE_NO_INIT)
  {
    UART_AdvFeatureConfig(huart);
  }

  /* In asynchronous mode, the following bits must be kept cleared:
  - LINEN and CLKEN bits in the USART_CR2 register,
  - SCEN, HDSEL and IREN  bits in the USART_CR3 register.*/
  CLEAR_BIT(huart->Instance->CR2, (USART_CR2_LINEN | USART_CR2_CLKEN));
  CLEAR_BIT(huart->Instance->CR3, (USART_CR3_SCEN | USART_CR3_HDSEL | USART_CR3_IREN));

  __HAL_UART_ENABLE(huart);

  /* TEACK and/or REACK to check before moving huart->gState and huart->RxState to Ready */
  return (UART_CheckIdleState(huart));
}



/**
  * @brief Send an amount of data in DMA mode.
  * @note   When UART parity is not enabled (PCE = 0), and Word Length is configured to 9 bits (M1-M0 = 01),
  *         the sent data is handled as a set of u16. In this case, Size must indicate the number
  *         of u16 provided through pData.
  * @param huart UART handle.
  * @param pData Pointer to data buffer (u8 or u16 data elements).
  * @param Size  Amount of data elements (u8 or u16) to be sent.
  * @retval HAL status
  */
HAL_StatusTypeDef HAL_UART_Transmit_DMA_rjg(UART_HandleTypeDef *huart, uint8_t *pData, uint16_t Size)
{
  /* Check that a Tx process is not already ongoing */
  if (huart->gState == HAL_UART_STATE_READY)
  {
    if ((pData == NULL) || (Size == 0U))
    {
      return HAL_ERROR;
    }

    __HAL_LOCK(huart);

    huart->pTxBuffPtr  = pData;
    huart->TxXferSize  = Size;
    huart->TxXferCount = Size;

    huart->ErrorCode = HAL_UART_ERROR_NONE;
    huart->gState = HAL_UART_STATE_BUSY_TX;

    if (huart->hdmatx != NULL)
    {
      /* Enable the UART transmit DMA channel */
      if (HAL_DMA_Start_IT(huart->hdmatx, (uint32_t)huart->pTxBuffPtr, (uint32_t)&huart->Instance->TDR, Size) != HAL_OK)
      {
        /* Set error code to DMA */
        huart->ErrorCode = HAL_UART_ERROR_DMA;

        __HAL_UNLOCK(huart);

        /* Restore huart->gState to ready */
        huart->gState = HAL_UART_STATE_READY;

        return HAL_ERROR;
      }
    }
    /* Clear the TC flag in the ICR register */
    __HAL_UART_CLEAR_FLAG(huart, UART_CLEAR_TCF);

    __HAL_UNLOCK(huart);

    /* Enable the DMA transfer for transmit request by setting the DMAT bit
    in the UART CR3 register */
    SET_BIT(huart->Instance->CR3, USART_CR3_DMAT);

    return HAL_OK;
  }
  else
  {
    return HAL_BUSY;
  }
}


/**
  * Uart 1 TX DMA Interrupt Handler.
  */
void DMA1_Stream7_IRQHandler(void)
{
	BaseType_t xHigherPriorityTaskWoken = pdFALSE;

  HAL_DMA_IRQHandler(&hdma_usart1_tx);

  vTaskNotifyGiveFromISR(printTaskHandle,
                       &xHigherPriorityTaskWoken);

  portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}


/**
  *Uart 1 RX DMA Interrupt Handler.
  */
void DMA1_Stream6_IRQHandler(void)
{
  HAL_DMA_IRQHandler(&hdma_usart1_rx);
}


/**
  * @brief This function handles USART1 global interrupt.
  */
void USART1_IRQHandler(void)
{
  HAL_UART_IRQHandler(&huart1);
}


/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
static void Error_Handler(void)
{
  /* User can add his own implementation to report the HAL error return state */
  __disable_irq();
  while (1)
  {
  }
}
